/*
 * Copyright 2012  Patrick Uiterwijk
 */
implementation module ArgumentParser

import StdEnv
import ArgEnv

:: ArgumentType = FlagArgument | StringValueArgument | IntValueArgument
:: ArgumentName :== String
:: ArgumentDescription :== String
:: Argument :== (ArgumentName, ArgumentType, ArgumentDescription)
:: ArgumentValue = ArgumentUnset | ArgumentSet | ArgumentStringValue String | ArgumentIntValue Int
:: ArgumentParserState = { programName :: String
			 , programDescription :: String
			 , validArguments :: [Argument]
			 , originalValidArguments :: [Argument]}

:: ParseResult a = Invalid ArgumentParserState | Parsed [(ArgumentName, ArgumentValue)]

initializeArgumentParser :: ArgumentParserState
initializeArgumentParser = { programName = ""
			   , programDescription = ""
			   , validArguments = []
			   , originalValidArguments = []}

setProgramName :: ArgumentParserState String -> ArgumentParserState
setProgramName state name = {state & programName = name}

setProgramDescription :: ArgumentParserState String -> ArgumentParserState
setProgramDescription state description = {state & programDescription = description}

addValidArgument :: ArgumentParserState Argument -> ArgumentParserState
addValidArgument state (newName, newType, newDesc) = {state & validArguments = [("-" +++ newName, newType, newDesc):state.validArguments]
							    , originalValidArguments = [("-" +++ newName, newType, newDesc):state.originalValidArguments]}

parseArguments :: ArgumentParserState -> (ArgumentName ->ArgumentValue)
parseArguments state = retrievalFunction (output (parse state (tl ([x \\ x <-: getCommandLine])) []))



retrievalFunction :: [(ArgumentName, ArgumentValue)] ArgumentName -> ArgumentValue
retrievalFunction [] reqName		= abort ("Unknown argument requested (" +++ reqName +++ ")\n")
retrievalFunction [(name, value):vs] reqName
| name == reqName			= value
| otherwise				= retrievalFunction vs reqName

parse :: ArgumentParserState [String] [(ArgumentName, ArgumentValue)] -> ParseResult [(ArgumentName, ArgumentValue)]
parse state [] untilNow				= Parsed (untilNow ++ (setRemainingAsUnset state.validArguments))
parse state [a:as] untilNow
| not (validArgument state.validArguments a)	= Invalid state
# argumentName = a	//For the time being. Should strip the first -
# argumentType = getArgumentType state.validArguments a
# state = {state & validArguments = (removeValidArgument state.validArguments argumentName)}
| argumentType == FlagArgument			= parse state as [(argumentName, ArgumentSet)]
| argumentType == StringValueArgument		= parse state (tl as) [(argumentName, ArgumentStringValue (toString (hd as))): untilNow]
| argumentType == IntValueArgument		= parse state (tl as) [(argumentName, ArgumentIntValue (toInt (hd as))): untilNow]



removeValidArgument :: [Argument] String -> [Argument]
removeValidArgument [] _	= []
removeValidArgument [(name, type, desc):as] pName
| name == pName			= removeValidArgument as pName
| otherwise			= [(name, type, desc): removeValidArgument as pName]

setRemainingAsUnset :: [Argument] -> [(ArgumentName, ArgumentValue)]
setRemainingAsUnset []			= []
setRemainingAsUnset [(name, _, _):as]	= [(name, ArgumentUnset): setRemainingAsUnset as]

validArgument :: [Argument] String -> Bool
validArgument [] _	= False
validArgument [(name, _, _):as] s
| name == s		= True
| otherwise		= validArgument as s

getArgumentType :: [Argument] String -> ArgumentType
getArgumentType [(name, type, _):as] s
| name == s		= type
| otherwise		= getArgumentType as s

output :: (ParseResult [(ArgumentName, ArgumentValue)]) -> [(ArgumentName, ArgumentValue)]
output (Invalid state)	= abort (showHelp state)
output (Parsed a)	= normalize a

normalize :: [(ArgumentName, ArgumentValue)] -> [(ArgumentName, ArgumentValue)]
normalize []			= []
normalize [(name, value):as]	= [(normalizeName name, value): normalize as]

normalizeName :: ArgumentName -> ArgumentName
normalizeName name = {a \\ a<-(tl [b \\ b <-: name])}

showHelp :: ArgumentParserState -> String
showHelp state = "Error: invalid usage.\n" +++ state.programName +++ ": " +++ state.programDescription +++ "\nValid options:\n" +++ (showHelpArgs state.originalValidArguments)

instance toString Argument
where
	toString :: !Argument -> {#Char}
	toString (name, StringValueArgument, desc)	= name +++ " [value] \t\t" +++ desc
	toString (name, IntValueArgument, desc)		= name +++ " [value] \t\t" +++ desc
	toString (name, type, desc)			= name +++ "\t\t\t" +++ desc

instance == ArgumentType
where
	(==) :: !ArgumentType !ArgumentType -> Bool
	(==) FlagArgument FlagArgument			= True
	(==) StringValueArgument StringValueArgument	= True
	(==) IntValueArgument IntValueArgument		= True
	(==) _ _					= False

instance toString ArgumentValue
where
	toString :: !ArgumentValue -> {#Char}
	toString ArgumentUnset		= "False"
	toString ArgumentSet		= "True"
	toString (ArgumentStringValue a)= a
	toString (ArgumentIntValue a)	= toString a

instance == ArgumentValue
where
	(==) :: !ArgumentValue !ArgumentValue -> Bool
	(==) ArgumentUnset ArgumentUnset			= True
	(==) ArgumentSet ArgumentSet				= True
	(==) (ArgumentStringValue a) (ArgumentStringValue b)	= a == b
	(==) (ArgumentIntValue a) (ArgumentIntValue b)		= a == b
	(==) _ _						= False

showHelpArgs :: [Argument] -> String
showHelpArgs []		= ""
showHelpArgs [a:as]	= toString a +++ "\n" +++ showHelpArgs as
